/*
    Scanbot SDK Flutter Plugin

    Copyright (c) 2019 doo GmbH. All rights reserved.
 */
package io.scanbot.sdk.flutter

import android.app.Activity
import com.google.gson.Gson
import io.flutter.plugin.common.MethodCall
import io.scanbot.sdk.barcode.entity.BarcodeFormat
import io.scanbot.sdk.barcode.entity.EngineMode
import io.scanbot.sdk.flutter.utils.ObjectMapper
import io.scanbot.sdk.ui.camera.FinderAspectRatio
import io.scanbot.sdk.ui.view.barcode.batch.configuration.BatchBarcodeScannerConfiguration
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeImageGenerationType
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeScannerAdditionalConfiguration
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeScannerConfiguration
import io.scanbot.sdk.ui.view.base.configuration.CameraOrientationMode
import io.scanbot.sdk.ui.view.base.configuration.OrientationMode
import io.scanbot.sdk.ui.view.camera.configuration.DocumentScannerConfiguration
import io.scanbot.sdk.ui.view.edit.configuration.CroppingConfiguration
import io.scanbot.sdk.ui.view.hic.configuration.HealthInsuranceCardScannerConfiguration
import io.scanbot.sdk.ui.view.licenseplate.configuration.LicensePlateScannerConfiguration
import io.scanbot.sdk.ui.view.mrz.configuration.MRZScannerConfiguration
import io.scanbot.sdk.ui.view.nfc.configuration.NfcPassportConfiguration


fun readEhicScannerConfiguration(
        call: MethodCall,
        activity: Activity
): HealthInsuranceCardScannerConfiguration {
    val cameraConfiguration = HealthInsuranceCardScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}

fun readLicensePlateScannerConfiguration(
        call: MethodCall,
        activity: Activity
): LicensePlateScannerConfiguration {
    val cameraConfiguration = LicensePlateScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}


fun readNFCPassportScannerConfiguration(
        call: MethodCall,
        activity: Activity
): NfcPassportConfiguration {
    val cameraConfiguration = NfcPassportConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}

fun readBarcodeScannerConfiguration(
        call: MethodCall,
        activity: Activity
): BarcodeScannerConfiguration {
    val barcodeCameraConfiguration = BarcodeScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        barcodeCameraConfiguration.setOrientationLockMode(value)
    }
    val aspectRatio = checkFinderAspectRatioParam(call, arguments)
    aspectRatio?.let {
        barcodeCameraConfiguration.setFinderAspectRatio(it)
    }
    ObjectMapper.map(arguments, barcodeCameraConfiguration)
    (arguments.get("additionalParameters") as? Json)?.let {
        barcodeCameraConfiguration.setAdditionalDetectionParameters(getAdditionalConfig(it))
    }
    if (arguments.containsKey("engineMode")) {
        val engineMode = arguments["engineMode"] as String
        val mode = EngineMode.valueOf(engineMode)
        barcodeCameraConfiguration.setEngineMode(mode)
    }
    if (arguments.containsKey("barcodeFormats")) {
        val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
        val nativeBarcodeFormats = ArrayList<BarcodeFormat>()
        for (formatValue in barcodeFormats) {
            nativeBarcodeFormats.add(BarcodeFormat.valueOf(formatValue.replaceFirst("BarcodeFormat.", "")))
        }
        barcodeCameraConfiguration.setBarcodeFormatsFilter(nativeBarcodeFormats)
    }
    if (arguments.containsKey("barcodeImageGenerationType")) {
        val imageGenType = BarcodeImageGenerationType.valueOf(arguments["barcodeImageGenerationType"] as String)
        barcodeCameraConfiguration.setBarcodeImageGenerationType(imageGenType)
    }
    return barcodeCameraConfiguration
}

private fun getAdditionalConfig(it: Json): BarcodeScannerAdditionalConfiguration {
    val defaultConfig = BarcodeScannerAdditionalConfiguration()
    val minimumTextLength = it["minimumTextLength"] as? Int
            ?: defaultConfig.minimumTextLength
    val enableGS1Decoding = it["enableGS1Decoding"] as? Boolean
            ?: defaultConfig.gs1DecodingEnabled
    val maximumTextLength = it["maximumTextLength"] as? Int
            ?: defaultConfig.maximumTextLength
    val minimum1DBarcodesQuietZone = it["minimum1DBarcodesQuietZone"] as? Int
            ?: defaultConfig.minimum1DQuietZoneSize
    return BarcodeScannerAdditionalConfiguration(
            gs1DecodingEnabled = enableGS1Decoding,
            minimumTextLength = minimumTextLength,
            maximumTextLength = maximumTextLength,
            minimum1DQuietZoneSize = minimum1DBarcodesQuietZone)
}

fun readBatchBarcodeScannerConfiguration(
        call: MethodCall,
        activity: Activity
): BatchBarcodeScannerConfiguration {
    val barcodeCameraConfiguration = BatchBarcodeScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        barcodeCameraConfiguration.setOrientationLockMode(value)
    }
    val aspectRatio = checkFinderAspectRatioParam(call, arguments)
    aspectRatio?.let {
        barcodeCameraConfiguration.setFinderAspectRatio(it)
    }
    ObjectMapper.map(arguments, barcodeCameraConfiguration)
    (arguments.get("additionalParameters") as? Json)?.let {
        barcodeCameraConfiguration.setAdditionalDetectionParameters(getAdditionalConfig(it))
    }
    if (arguments.containsKey("engineMode")) {
        val engineMode = arguments["engineMode"] as String
        val mode = EngineMode.valueOf(engineMode)
        barcodeCameraConfiguration.setEngineMode(mode)
    }
    if (arguments.containsKey("barcodeFormats")) {
        val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
        val nativeBarcodeFormats = ArrayList<BarcodeFormat>()
        for (formatValue in barcodeFormats) {
            nativeBarcodeFormats.add(BarcodeFormat.valueOf(formatValue.replaceFirst("BarcodeFormat.", "")))
        }
        barcodeCameraConfiguration.setBarcodeFormatsFilter(nativeBarcodeFormats)
    }

    return barcodeCameraConfiguration
}

fun readCroppingConfiguration(
        call: MethodCall,
        activity: Activity,
        gson: Gson
): CroppingConfiguration {
    val croppingConfiguration = CroppingConfiguration()
    val arguments = call.arguments as Json
    call.argument<Json>("page")?.let {
        val page = it.toPage(gson = gson)
        croppingConfiguration.setPage(page)
        arguments.remove("page")
    }
    val value = checkOrientationParam(call, arguments)
    value?.let {
        croppingConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, croppingConfiguration)
    return croppingConfiguration
}

fun readMRZScannerConfiguration(
        call: MethodCall,
        activity: Activity
): MRZScannerConfiguration {
    val cameraConfiguration = MRZScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    val aspectRatio = checkFinderAspectRatioParam(call, arguments)
    aspectRatio?.let {
        cameraConfiguration.setFinderAspectRatio(it)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}

private fun checkCameraOrientationParam(call: MethodCall, arguments: Json): CameraOrientationMode? {
    if (call.hasArgument("orientationLockMode")) {
        val value = when (call.argument<String>("orientationLockMode")) {
            "PORTRAIT" -> CameraOrientationMode.PORTRAIT
            "PORTRAIT_UPSIDE_DOWN" -> CameraOrientationMode.PORTRAIT
            "LANDSCAPE_LEFT" -> CameraOrientationMode.LANDSCAPE
            "LANDSCAPE_RIGHT" -> CameraOrientationMode.LANDSCAPE
            "LANDSCAPE" -> CameraOrientationMode.LANDSCAPE
            else -> CameraOrientationMode.PORTRAIT
        }
        arguments.remove("orientationLockMode")
        return value
    }
    return null
}

private fun checkOrientationParam(call: MethodCall, arguments: Json): OrientationMode? {
    if (call.hasArgument("orientationLockMode")) {
        val value = when (call.argument<String>("orientationLockMode")) {
            "PORTRAIT" -> OrientationMode.PORTRAIT
            "PORTRAIT_UPSIDE_DOWN" -> OrientationMode.PORTRAIT
            "LANDSCAPE_LEFT" -> OrientationMode.LANDSCAPE
            "LANDSCAPE_RIGHT" -> OrientationMode.LANDSCAPE
            "LANDSCAPE" -> OrientationMode.LANDSCAPE
            else -> OrientationMode.PORTRAIT
        }
        arguments.remove("orientationLockMode")
        return value
    }
    return null
}


private fun checkFinderAspectRatioParam(call: MethodCall, arguments: Json): FinderAspectRatio? {
    val propertyName = "finderAspectRatio"
    if (arguments[propertyName] != null) {
        val aspRatio = arguments[propertyName] as? Json
        val width = aspRatio?.get("width") as? Int
        val height = aspRatio?.get("height") as? Int
        arguments.remove(propertyName)
        return FinderAspectRatio(width?.toDouble() ?: 1.0, height?.toDouble() ?: 1.0)
    }
    return null
}

fun readDocumentScannerConfiguration(
        call: MethodCall,
        activity: Activity
): DocumentScannerConfiguration {
    val cameraConfiguration = DocumentScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}


