//
//  SDKCallHandler.swift
//  Runner
//
//  Created by Zhenya Knizhnik on 22.08.19.
//  Copyright © 2019 The Chromium Authors. All rights reserved.
//

import Foundation
import Flutter
import ScanbotSDK

typealias CallParameters = (call: FlutterMethodCall, result: FlutterResult)
typealias CallUIParameters = (call: FlutterMethodCall, result: FlutterResult, presenter: UIViewController?)

class SDKCallHandler: CallErrorHandler {
    let rtuuiPresenter = RTUUIPresenter()
    var dataFormatterChannel: FlutterMethodChannel?

    func handle(call: FlutterMethodCall, rootViewController: UIViewController?, result: @escaping FlutterResult) {
        
        let uiParameters: CallUIParameters = (call: call, result: result, presenter: rootViewController)
        let parameters: CallParameters = (call: call, result: result)

        switch call.method {
        case "initScanbotSdk": self.initSDK(parameters)
        case "getLicenseStatus": self.getLicenseStatus(result: result)
        case "cleanupStorage": self.cleanupStorage(result: result)
        case "rotatePageClockwise": self.rotatePage(parameters)
        case "deletePage": self.deletePage(parameters)
        case "applyImageFilter": self.applyImageFilter(parameters)
        case "getFilteredDocumentPreviewUri": self.getFilteredDocumentPreviewURL(parameters)
        case "createPdf": self.createPDF(parameters)
        case "createTiff": self.createTIFF(parameters)
        case "createPage": self.createPage(parameters)
        case "detectDocument": self.detectDocument(parameters)
        case "performOcr": self.performOCR(parameters)
        case "estimateBlurOnPage": self.estimateBlur(parameters)
        case "getOcrConfig": self.getOCRConfig(result: result)
        case "refreshImageUris": self.refreshImageURLS(call: call, result: result)
        case "detectBarcodesFromImageFile": self.detectBarcodesFromImageFile(parameters)
        case "getDecryptedDataFromFile": self.getDecryptFileFromURL(parameters)
            
        case "startDocumentScanner": self.rtuuiPresenter.showDocumentScanner(uiParameters)
        case "startBarcodeScanner": self.rtuuiPresenter.showBarcodeScanner(uiParameters)
        case "startMrzScanner": self.rtuuiPresenter.showMRZScanner(uiParameters)
        case "startEhicScanner": self.rtuuiPresenter.showEHICScanner(uiParameters)
        case "startCroppingScreen": self.rtuuiPresenter.showCropScreen(uiParameters)
        case "startBatchBarcodeScanner": self.rtuuiPresenter.showBatchScanner(uiParameters,
                                                                              dataFormatterChannel: self.dataFormatterChannel)
        case "startLicensePlateScanner": self.rtuuiPresenter.showLicensePlateScanner(uiParameters)
        case "startNfcPassportReader": self.rtuuiPresenter.showNFCReader(uiParameters)
            
        default: self.returnUnknownCallError(call: call, result: result)
        }
    }
    
    private func getLicenseStatus(result: @escaping FlutterResult) {
        let date = ScanbotSDK.licenseExpirationDate()
        let isValid = ScanbotSDK.isLicenseValid()
        let status = ScanbotSDK.licenseStatus()
        
        let statusResult = LicenseStatus(isLicenseValid: isValid, status: status, expirationDate: date)
        result(statusResult.encodedString)
    }
    
    private func initSDK(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let config = SDKInitializationConfig(from: arguments) else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        ScanbotSDKInitializer.initialize(config: config, result: result)
    }
    
    private func getDecryptFileFromURL(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
              let fileURLString = (arguments["fileUri"] as? String)?.removingPercentEncoding,
              let dataToDecrypt = try? Data(contentsOf: URL(fileURLWithPath: fileURLString)) else {
            
                self.returnArgumentsError(call: call, result: result)
                return
        }
        
        // We are using imageStorage encryptor, as the encryptor should be the same for all cases.
        if let encrypter = ScanbotSDKUI.defaultImageStoreEncrypter() {
            let decryptedData = encrypter.decryptData(dataToDecrypt)
            result(decryptedData)
        } else {
            // To mimic Android SDK behavior we are just returning the data of the file, in case no encrypter  is set.
            result(dataToDecrypt)
        }
    }
    
    private func cleanupStorage(result: @escaping FlutterResult) {
        SBSDKUIPageFileStorage.default().removeAll()
        
        if let url = PluginPreferences.shared.storageCustomFilesURL {
            self.clearDirectory(directoryURL: url)
        }
        
        result(nil)
    }
    
    private func clearDirectory(directoryURL: URL) {
        let fileManager = FileManager.default
        do {
            let urls = try fileManager.contentsOfDirectory(at: directoryURL,
                                                           includingPropertiesForKeys: nil,
                                                           options: FileManager.DirectoryEnumerationOptions())
            for url in urls {
                try fileManager.removeItem(at: url)
            }
        } catch {
            print("Could not clear folder: \(error)")
        }
    }
    
    private func getOCRConfig(result: @escaping FlutterResult) {
        if let config = OCRConfigGetter.encodedOCRConfig {
            result(config)
        } else {
            result(FlutterError(code: "1011", message: "ERROR_iOS: OCRConfig getting error", details: nil))
        }
    }
    
    private func detectDocument(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let page = SBSDKUIPage(from: arguments),
            let image = page.originalImage() else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        
        let polygon = DocumentDetector().detectPolygon(on: image)
        page.polygon = polygon
        
        result(page.encodedString)
    }
    
    private func performOCR(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let config = OCRPerformConfiguration(from: arguments) else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        OCRPerformer().perform(with: config, result: result)
    }
    
    private func refreshImageURLS(call: FlutterMethodCall, result: @escaping FlutterResult) {
        // As the location of the Application Support folder can change (e.g. on app update), we simply recreate
        // Page objects here to get a refreshed absolute file paths of the images.
        // (under the hood the native iOS SDK works with relative paths and is able to handle variable Application Support location)
        guard let arguments = call.arguments as? [String: Any]  else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        let pages = SBSDKUIPage.pages(from: arguments)
        let dictionary = ["pages": pages.map({$0.encodedDictionary})]
        result(dictionary.encodedString)
    }
    
    private func estimateBlur(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let page = SBSDKUIPage(from: arguments),
            let image = page.documentImage() else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        
        let value = SBSDKBlurrinessEstimator().estimateImageBlurriness(image)
        result(value)
    }
    
    private func createPDF(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let pageSize = SBSDKPDFRendererPageSize(from: arguments) else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        let pages = SBSDKUIPage.pages(from: arguments)
        Renderer().renderPDF(from: pages, pageSize: pageSize, result: result)
    }
    
    private func createTIFF(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any]  else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        let options = TIFFOptions(from: arguments)
        let pages = SBSDKUIPage.pages(from: arguments)
        Renderer().renderTIFF(from: pages, options: options, result: result)
    }
    
    private func createPage(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        DispatchQueue.main.async {
            guard let arguments = call.arguments as? [String: Any],
                let imageURLString = arguments["originalImageFileUri"] as? String,
                let shouldDetect = arguments["shouldDetectDocument"] as? Bool  else {
                    self.returnArgumentsError(call: call, result: result)
                    return
            }
            
            guard let imageURL = URL(string: imageURLString) else {
                self.returnArgumentsError(call: call, result: result)
                return
            }
            
            guard let imageData = try? Data(contentsOf: imageURL),
                let image = UIImage(data: imageData)?.sbsdk_imageWithNormalizedOrientation(),
                let uuid = SBSDKUIPageFileStorage.default().add(image) else {
                    self.returnArgumentsError(call: call, result: result)
                    return
            }
            
            var polygon: SBSDKPolygon?
            if shouldDetect {
                polygon = DocumentDetector().detectPolygon(on: image)
            }
            
            let page = SBSDKUIPage(pageFileID: uuid, polygon: polygon)
            result(page.encodedString)
        }
    }

    
    private func rotatePage(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let page = SBSDKUIPage(from: arguments),
            let count = arguments["rotationsCount"] as? Int else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        page.rotateClockwise(count)
        result(page.encodedString)
    }
    
    private func applyImageFilter(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let page = SBSDKUIPage(from: arguments),
            let filter = SBSDKImageFilterType(from: arguments) else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        page.filter = filter
        result(page.encodedString)
    }
    
    private func getFilteredDocumentPreviewURL(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let page = SBSDKUIPage(from: arguments),
            let filter = SBSDKImageFilterType(from: arguments) else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        let url = page.documentPreviewImageURL(using: filter)
        result(url?.absoluteString)
    }
    
    private func deletePage(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let page = SBSDKUIPage(from: arguments) else {
                self.returnArgumentsError(call: call, result: result)
                return
        }
        SBSDKUIPageFileStorage.default().removePageFileID(page.pageFileUUID)
        result(nil)
    }
    
    private func detectBarcodesFromImageFile(_ parameters: CallParameters) {
        let (call, result) = (parameters.call, parameters.result)

        guard let arguments = call.arguments as? [String: Any],
            let inputPath = arguments["uri"] as? String,
            let url = URL(string: inputPath)?.path,
            let image = UIImage(contentsOfFile: url) else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        
        let barcodeTypes = SBSDKBarcodeType.fromArguments(arguments) ?? []
        
        let detector = SBSDKBarcodeScanner(types: barcodeTypes)
        let barcodeResults = detector.detectBarCodes(on: image) ?? []
        let barcodeList = barcodeResults.map({ $0.encodedDictionary })
        let encodedResult: [String: Any?] = ["barcodeItems": barcodeList]

        result(encodedResult.encodedString)
    }
    
}
